@@ -1,7 +1,6 @@
 /* Parsing FTP `ls' output.
-   Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
-   2005, 2006, 2007, 2008, 2009, 2010, 2011, 2015 Free Software
-   Foundation, Inc.
+   Copyright (C) 1996-2011, 2015, 2018-2023 Free Software Foundation,
+   Inc.
 
 This file is part of GNU Wget.
 
@@ -65,7 +64,7 @@ symperms (const char *s)
 
 
 /* Cleans a line of text so that it can be consistently parsed. Destroys
-   <CR> and <LF> in case that thay occur at the end of the line and
+   <CR> and <LF> in case that they occur at the end of the line and
    replaces all <TAB> character with <SPACE>. Returns the length of the
    modified line. */
 static int
@@ -89,12 +88,11 @@ clean_line (char *line, int len)
    output (whatever that might be).  BSD (no group) and SYSV (with
    group) listings are handled.
 
-   The time stamps are stored in a separate variable, time_t
+   The timestamps are stored in a separate variable, time_t
    compatible (I hope).  The timezones are ignored.  */
 static struct fileinfo *
-ftp_parse_unix_ls (const char *file, int ignore_perms)
+ftp_parse_unix_ls (FILE *fp, int ignore_perms)
 {
-  FILE *fp;
   static const char *months[] = {
     "Jan", "Feb", "Mar", "Apr", "May", "Jun",
     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
@@ -109,12 +107,6 @@ ftp_parse_unix_ls (const char *file, int ignore_perms)
   char *line = NULL, *tok, *ptok;      /* tokenizer */
   struct fileinfo *dir, *l, cur;       /* list creation */
 
-  fp = fopen (file, "rb");
-  if (!fp)
-    {
-      logprintf (LOG_NOTQUIET, "%s: %s\n", file, strerror (errno));
-      return NULL;
-    }
   dir = l = NULL;
 
   /* Line loop to end of file: */
@@ -167,12 +159,12 @@ ftp_parse_unix_ls (const char *file, int ignore_perms)
               /*cur.perms = 1023;*/     /* #### What is this?  --hniksic */
               cur.perms = 0644;
             }
-          DEBUGP (("implicit perms %0o; ", cur.perms));
+          DEBUGP (("implicit perms %0o; ", (unsigned) cur.perms));
         }
        else
          {
            cur.perms = symperms (tok + 1);
-           DEBUGP (("perms %0o; ", cur.perms));
+           DEBUGP (("perms %0o; ", (unsigned) cur.perms));
          }
 
       error = ignore = 0;       /* Erroneous and ignoring entries are
@@ -251,24 +243,25 @@ ftp_parse_unix_ls (const char *file, int ignore_perms)
               /* We must deal with digits.  */
               if (c_isdigit (*tok))
                 {
-                  /* Suppose it's year.  */
-                  for (; c_isdigit (*tok); tok++)
+                  /* Suppose it's year.  Limit to year 99999 to avoid integer overflow. */
+                  for (; c_isdigit (*tok) && year <= 99999; tok++)
                     year = (*tok - '0') + 10 * year;
                   if (*tok == ':')
                     {
+                      int n;
                       /* This means these were hours!  */
                       hour = year;
                       year = 0;
                       ptype = TT_HOUR_MIN;
                       ++tok;
                       /* Get the minutes...  */
-                      for (; c_isdigit (*tok); tok++)
+                      for (n = 0; c_isdigit (*tok) && n < 2; tok++, n++)
                         min = (*tok - '0') + 10 * min;
                       if (*tok == ':')
                         {
                           /* ...and the seconds.  */
                           ++tok;
-                          for (; c_isdigit (*tok); tok++)
+                          for (n = 0; c_isdigit (*tok) && n < 2; tok++, n++)
                             sec = (*tok - '0') + 10 * sec;
                         }
                     }
@@ -415,14 +408,12 @@ ftp_parse_unix_ls (const char *file, int ignore_perms)
     }
 
   xfree (line);
-  fclose (fp);
   return dir;
 }
 
 static struct fileinfo *
-ftp_parse_winnt_ls (const char *file)
+ftp_parse_winnt_ls (FILE *fp)
 {
-  FILE *fp;
   int len;
   int year, month, day;         /* for time analysis */
   int hour, min;
@@ -433,13 +424,8 @@ ftp_parse_winnt_ls (const char *file)
   char *filename;
   struct fileinfo *dir, *l, cur; /* list creation */
 
-  fp = fopen (file, "rb");
-  if (!fp)
-    {
-      logprintf (LOG_NOTQUIET, "%s: %s\n", file, strerror (errno));
-      return NULL;
-    }
   dir = l = NULL;
+  cur.name = NULL;
 
   /* Line loop to end of file: */
   while ((len = getline (&line, &bufsize, fp)) > 0)
@@ -458,8 +444,8 @@ ftp_parse_winnt_ls (const char *file)
          january will be assumed.  */
       tok = strtok(line, "-");
       if (tok == NULL) continue;
-      month = atoi(tok) - 1;
-      if (month < 0) month = 0;
+      month = atoi(tok);
+      if (month < 0) month = 0; else month--;
       tok = strtok(NULL, "-");
       if (tok == NULL) continue;
       day = atoi(tok);
@@ -474,10 +460,13 @@ ftp_parse_winnt_ls (const char *file)
       else if (year >= 1900)
         {
           year -= 1900;
+          if (len < 42) continue;
           filename += 2;
         }
       /* Now it is possible to determine the position of the first symbol in
          filename. */
+      xfree (cur.name);
+      memset(&cur, 0, sizeof (cur));
       cur.name = xstrdup(filename);
       DEBUGP (("Name: '%s'\n", cur.name));
 
@@ -492,8 +481,8 @@ ftp_parse_winnt_ls (const char *file)
       min = atoi(tok);
       /* Adjust hour from AM/PM. Just for the record, the sequence goes
          11:00AM, 12:00PM, 01:00PM ... 11:00PM, 12:00AM, 01:00AM . */
-      tok+=2;
-      if (hour == 12)  hour  = 0;
+      if (tok[0] && tok[1]) tok+=2;
+      if (hour >= 12 || hour < 0)  hour  = 0;
       if (*tok == 'P') hour += 12;
 
       DEBUGP (("YYYY/MM/DD HH:MM - %d/%02d/%02d %02d:%02d\n",
@@ -560,10 +549,11 @@ ftp_parse_winnt_ls (const char *file)
           memcpy (l, &cur, sizeof (cur));
           l->next = NULL;
         }
+      cur.name = NULL;
     }
 
+  xfree (cur.name);
   xfree (line);
-  fclose(fp);
   return dir;
 }
 
@@ -605,7 +595,6 @@ static void eat_carets( char *str)
   char *strd;   /* Destination pointer. */
   char hdgt;
   unsigned char uchr;
-  unsigned char prop;
 
   /* Skip ahead to the first "^", if any. */
   while ((*str != '\0') && (*str != '^'))
@@ -622,28 +611,27 @@ static void eat_carets( char *str)
       if (uchr == '^')
       {
         /* Found a caret.  Skip it, and check the next character. */
-        uchr = *(++str);
-        prop = char_prop[ uchr];
-        if (prop& 64)
+        if ((char_prop[(unsigned char) str[1]] & 64) && (char_prop[(unsigned char) str[2]] & 64))
         {
           /* Hex digit.  Get char code from this and next hex digit. */
+          uchr = *(++str);
           if (uchr <= '9')
           {
-            hdgt = uchr- '0';           /* '0' - '9' -> 0 - 9. */
+            hdgt = uchr - '0';           /* '0' - '9' -> 0 - 9. */
           }
           else
           {
-            hdgt = ((uchr- 'A')& 7)+ 10;    /* [Aa] - [Ff] -> 10 - 15. */
+            hdgt = ((uchr - 'A') & 7) + 10;    /* [Aa] - [Ff] -> 10 - 15. */
           }
           hdgt <<= 4;                   /* X16. */
           uchr = *(++str);              /* Next char must be hex digit. */
           if (uchr <= '9')
           {
-            uchr = hdgt+ uchr- '0';
+            uchr = hdgt + uchr - '0';
           }
           else
           {
-            uchr = hdgt+ ((uchr- 'A')& 15)+ 10;
+            uchr = hdgt + ((uchr - 'A') & 15) + 10;
           }
         }
         else if (uchr == '_')
@@ -678,25 +666,18 @@ static void eat_carets( char *str)
 
 
 static struct fileinfo *
-ftp_parse_vms_ls (const char *file)
+ftp_parse_vms_ls (FILE *fp)
 {
-  FILE *fp;
   int dt, i, j, len;
   int perms;
   size_t bufsize = 0;
   time_t timenow;
   struct tm *timestruct;
-  char date_str[ 32];
+  char date_str[32];
 
   char *line = NULL, *tok; /* tokenizer */
   struct fileinfo *dir, *l, cur; /* list creation */
 
-  fp = fopen (file, "r");
-  if (!fp)
-    {
-      logprintf (LOG_NOTQUIET, "%s: %s\n", file, strerror (errno));
-      return NULL;
-    }
   dir = l = NULL;
 
   /* Skip blank lines, Directory heading, and more blank lines. */
@@ -728,6 +709,7 @@ ftp_parse_vms_ls (const char *file)
 
   /* Read remainder of file until the next blank line or EOF. */
 
+  cur.name = NULL;
   while (i > 0)
     {
       char *p;
@@ -754,8 +736,8 @@ ftp_parse_vms_ls (const char *file)
       */
 
 #if (!defined( __VMS) && !defined( PRESERVE_VMS_VERSIONS))
-      for (p = tok + strlen (tok); (--p > tok) && c_isdigit( *p); );
-      if ((*p == ';') && (*(p- 1) != '^'))
+      for (p = tok + strlen (tok); (--p > tok) && c_isdigit(*p); );
+      if (p > tok && (*p == ';') && (*(p - 1) != '^'))
         {
           *p = '\0';
         }
@@ -776,16 +758,16 @@ ftp_parse_vms_ls (const char *file)
          what will work in a CWD command.
       */
       len = strlen (tok);
-      if (!c_strncasecmp((tok + (len - 4)), ".DIR", 4))
+      if (len >= 4 && !c_strncasecmp(tok + (len - 4), ".DIR", 4))
         {
-          *(tok+ (len - 4)) = '\0'; /* Discard ".DIR". */
+          *(tok + (len - 4)) = '\0'; /* Discard ".DIR". */
           cur.type  = FT_DIRECTORY;
           cur.perms = VMS_DEFAULT_PROT_DIR;
           DEBUGP (("Directory (nv)\n"));
         }
-      else if (!c_strncasecmp ((tok + (len - 6)), ".DIR;1", 6))
+      else if (len >= 6 && !c_strncasecmp (tok + len - 6, ".DIR;1", 6))
         {
-          *(tok+ (len - 6)) = '\0'; /* Discard ".DIR;1". */
+          *(tok + (len - 6)) = '\0'; /* Discard ".DIR;1". */
           cur.type  = FT_DIRECTORY;
           cur.perms = VMS_DEFAULT_PROT_DIR;
           DEBUGP (("Directory (v)\n"));
@@ -796,6 +778,7 @@ ftp_parse_vms_ls (const char *file)
           cur.perms = VMS_DEFAULT_PROT_FILE;
           DEBUGP (("File\n"));
         }
+      xfree (cur.name);
       cur.name = xstrdup (tok);
       DEBUGP (("Name: '%s'\n", cur.name));
 
@@ -874,8 +857,7 @@ ftp_parse_vms_ls (const char *file)
             {
               /* Date. */
               DEBUGP (("Date.\n"));
-              strcpy( date_str, tok);
-              strcat( date_str, " ");
+				  snprintf(date_str, sizeof(date_str), "%s ", tok);
             }
           else if ((strlen (tok) < 12) && (strchr( tok, ':') != NULL))
             {
@@ -909,12 +891,12 @@ ftp_parse_vms_ls (const char *file)
                       if (j == 0)
                         {
                           perms = 0;
-                          j = 1;
                         }
-                      else
+                      else if (j < 4)
                         {
                           perms <<= 3;
                         }
+                      j++;
                       break;
                     case 'R':
                       perms |= 4;
@@ -931,7 +913,7 @@ ftp_parse_vms_ls (const char *file)
                     }
                 }
               cur.perms = perms;
-              DEBUGP (("Prot.  perms = %0o.\n", cur.perms));
+              DEBUGP (("Prot.  perms = %0o.\n", (unsigned) cur.perms));
             }
           else
             {
@@ -978,18 +960,19 @@ ftp_parse_vms_ls (const char *file)
       /* Add the data for this item to the linked list, */
       if (!dir)
         {
-          l = dir = (struct fileinfo *)xmalloc (sizeof (struct fileinfo));
+          l = dir = xmalloc (sizeof (struct fileinfo));
+          cur.prev = cur.next = NULL;
           memcpy (l, &cur, sizeof (cur));
-          l->prev = l->next = NULL;
         }
       else
         {
           cur.prev = l;
-          l->next = (struct fileinfo *)xmalloc (sizeof (struct fileinfo));
+			 cur.next = NULL;
+          l->next = xmalloc (sizeof (struct fileinfo));
           l = l->next;
           memcpy (l, &cur, sizeof (cur));
-          l->next = NULL;
         }
+      cur.name = NULL;
 
       i = getline (&line, &bufsize, fp);
       if (i > 0)
@@ -1003,53 +986,65 @@ ftp_parse_vms_ls (const char *file)
         }
     }
 
+  xfree (cur.name);
   xfree (line);
-  fclose (fp);
   return dir;
 }
 
 
 /* This function switches between the correct parsing routine depending on
    the SYSTEM_TYPE. The system type should be based on the result of the
-   "SYST" response of the FTP server. According to this repsonse we will
+   "SYST" response of the FTP server. According to this response we will
    use on of the three different listing parsers that cover the most of FTP
    servers used nowadays.  */
 
 struct fileinfo *
 ftp_parse_ls (const char *file, const enum stype system_type)
+{
+  FILE *fp;
+  struct fileinfo *fi;
+
+  fp = fopen (file, "rb");
+  if (!fp)
+    {
+      logprintf (LOG_NOTQUIET, "%s: %s\n", file, strerror (errno));
+      return NULL;
+    }
+
+  fi = ftp_parse_ls_fp (fp, system_type);
+  fclose(fp);
+
+  return fi;
+}
+
+struct fileinfo *
+ftp_parse_ls_fp (FILE *fp, const enum stype system_type)
 {
   switch (system_type)
     {
     case ST_UNIX:
-      return ftp_parse_unix_ls (file, 0);
+      return ftp_parse_unix_ls (fp, 0);
     case ST_WINNT:
       {
         /* Detect whether the listing is simulating the UNIX format */
-        FILE *fp;
-        int   c;
-        fp = fopen (file, "rb");
-        if (!fp)
-        {
-          logprintf (LOG_NOTQUIET, "%s: %s\n", file, strerror (errno));
-          return NULL;
-        }
-        c = fgetc(fp);
-        fclose(fp);
+        int   c = fgetc(fp);
+        rewind(fp);
+
         /* If the first character of the file is '0'-'9', it's WINNT
            format. */
         if (c >= '0' && c <='9')
-          return ftp_parse_winnt_ls (file);
+          return ftp_parse_winnt_ls (fp);
         else
-          return ftp_parse_unix_ls (file, 1);
+          return ftp_parse_unix_ls (fp, 1);
       }
     case ST_VMS:
-      return ftp_parse_vms_ls (file);
+      return ftp_parse_vms_ls (fp);
     case ST_MACOS:
-      return ftp_parse_unix_ls (file, 1);
+      return ftp_parse_unix_ls (fp, 1);
     default:
       logprintf (LOG_NOTQUIET, _("\
 Unsupported listing type, trying Unix listing parser.\n"));
-      return ftp_parse_unix_ls (file, 0);
+      return ftp_parse_unix_ls (fp, 0);
     }
 }
 
